/*
 * memo.c --
 *
 *	Definition for packer wrappers to memo databases.
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: memo.c,v 1.1 1999/03/27 13:26:34 aku Exp $
 */

#include "wrap.h"
#include "category.h"
#include "internal.h"

#include "pi-memo.h"

typedef struct Memo            Memo;
typedef struct MemoAppInfo     MemoAppInfo;


/*
 * Definitions for conversion of memo records.
 */

static TclUtil_StructItem recordItems [] = {
  ITEM ("text", text, Memo, TUSI_String),
  ITEM_END
};

static TclUtil_Struct recordStruct = {
  "memo", 0, NULL, recordItems
};

static PiTcl_CvtData recordConverter = {
  &recordStruct,                  /* Definition of memo structure, for
				   * conversion to tcl list */
  sizeof (Memo),                  /* Size of the memo structure */
  (PiTcl_FreeStruct*) free_Memo,  /* Proc to free additionally allocated
				   * space */
  (PiTcl_CvtStruct*) pack_Memo,   /* Serialize memo into bytestream */
  (PiTcl_CvtStruct*) unpack_Memo, /* Unserialize bytestream into a nemo */
  4096                            /* Initial size of buffer to use for
				   * serialization */
};


/*
 * Definitions for conversion of memo application information.
 */

static TclUtil_StructItem appinfoItems [] = {
  ITEM_STRUCT ("category",    category,    MemoAppInfo, &catStruct),
  ITEM        ("sortByAlpha", sortByAlpha, MemoAppInfo, TUSI_Boolean),
  ITEM_END
};

static TclUtil_Struct appinfoStruct = {
  "memo application information", 0, NULL, appinfoItems
};

static PiTcl_CvtData appinfoConverter = {
  &appinfoStruct,                        /* Definition of memo appinfo
					  * structure, for conversion to and
					  * from a tcl list */
  sizeof (MemoAppInfo),                  /* Size of the memo appinfo structure
					  */
  NULL,                                  /* *No* additionally allocated space
					  */
  (PiTcl_CvtStruct*) pack_MemoAppInfo,   /* Serialize memo appinfo into a
					  * bytestream */
  (PiTcl_CvtStruct*) unpack_MemoAppInfo, /* Unserialize bytestream into a memo
					  * appinfo block */
  sizeof (MemoAppInfo)+1                 /* Initial size of buffer to use for
					  * serialization */
};

/*
 * Memo databases do not have a sort information block.
 */

/*
 * Definition of the whole memo converter.
 */

static PiTcl_PilotCvt converter = {
  &recordConverter,
  NULL,
  &appinfoConverter
};

/*
 *------------------------------------------------------*
 *
 *	PiTclInitMemo --
 *
 *	------------------------------------------------*
 *	This procedure registers the commands wrapped
 *	around the (un)pack functionality for memos.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclInitMemo (interp)
Tcl_Interp* interp;
{
  Tcl_CreateObjCommand (interp, "piMemo",
			&PiTclPilotCvtCmd,
			(ClientData) &converter,
			NULL);

  return TCL_OK;
}

